import csv
from datetime import datetime
import dateutil.tz
import ics.attendee
import ics.event
import nytid.schedules
import nytid.signup.utils
from nytid.signup import sheets
import re
import requests

SIGNUP_SHEET_HEADER = ["Event", "Start", "End", "Rooms", "#Needed TAs"]
DATETIME_FORMAT = "YYYY-MM-DD HH:mm"
STRPTIME_FORMAT = "%Y-%m-%d %H:%M"


class EventFromCSV(ics.event.Event):
    """A class to create an ics.event.Event from an event in CSV format"""

    def __init__(self, csv_row):
        """
        Input: a row from a calendar in CSV format (e.g. the sign-up sheet).
        """
        kwargs = dict()

        kwargs["name"] = csv_row[SIGNUP_SHEET_HEADER.index("Event")]

        tz = dateutil.tz.tzlocal()

        begin = datetime.strptime(
            csv_row[SIGNUP_SHEET_HEADER.index("Start")], STRPTIME_FORMAT
        )
        begin.replace(tzinfo=tz)
        kwargs["begin"] = begin.astimezone(dateutil.tz.UTC)

        end = datetime.strptime(
            csv_row[SIGNUP_SHEET_HEADER.index("End")], STRPTIME_FORMAT
        )
        end.replace(tzinfo=tz)
        kwargs["end"] = end.astimezone(dateutil.tz.UTC)

        rooms = csv_row[SIGNUP_SHEET_HEADER.index("Rooms")]
        kwargs["location"] = rooms
        needed_TAs = csv_row[SIGNUP_SHEET_HEADER.index("#Needed TAs")]
        kwargs["description"] = f"Needed TAs: {needed_TAs}\n"

        kwargs["attendees"] = [
            ics.attendee.Attendee(f"{user}@kth.se")
            for user in csv_row[len(SIGNUP_SHEET_HEADER) :]
            if user
        ]
        booked, reserves = sheets.get_booked_TAs_from_csv(csv_row)

        kwargs["description"] += "Booked TAs:"
        for username in booked:
            kwargs["description"] += f" {username}"
        kwargs["description"] += "\n"

        kwargs["description"] += "Reserve TAs:"
        for username in reserves:
            kwargs["description"] += f" {username}"
        kwargs["description"] += "\n"

        super().__init__(**kwargs)


def generate_signup_sheet(
    outfile,
    url,
    needed_TAs=nytid.signup.utils.needed_TAs,
    event_filter=nytid.schedules.event_filter,
):
    """
    Input:
    - outfile is a string containing the file name used for output.
    - url is the URL to the ICS-formatted calendar.
    - needed_TAs is a function computing the number of needed TAs based on the
      event. The default is the needed_TAs function in this module,
    - event_filter is a function that filters events, takes a list of events as
      argument and returns a filtered list.

    Output:
    Returns nothing. Writes output to {outfile}.csv.
    """
    with open(outfile, "w") as out:
        csvout = csv.writer(out, delimiter="\t")
        calendar = nytid.schedules.read_calendar(url)

        max_num_TAs = 0
        rows = []

        for event in event_filter(calendar.timeline):
            num_TAs = needed_TAs(event)

            if num_TAs > max_num_TAs:
                max_num_TAs = num_TAs

            rows.append(
                [
                    event.name,
                    event.begin.to("local").format(DATETIME_FORMAT),
                    event.end.to("local").format(DATETIME_FORMAT),
                    event.location,
                    num_TAs,
                ]
            )

        csvout.writerow(
            SIGNUP_SHEET_HEADER + [f"TA username" for n in range(max_num_TAs)] + ["..."]
        )

        csvout.writerows(rows)


def read_signup_sheet_from_file(filename):
    """
    Input: filename is a string containing the file name of the CSV file of the
    sign-up sheet.

    Output: All the rows of the CSV as a Python list.
    """
    with open(filename, "r") as f:
        csvfile = csv.reader(f)
        return list(filter(any, list(csvfile)[1:]))


def read_signup_sheet_from_url(url):
    """
    Input: url is a string containing the URL of the CSV file of the sign-up
    sheet.

    Output: All the rows of the CSV as a Python list.
    """
    response = requests.get(url)
    if response.status_code != 200:
        raise ValueError(response.text)

    response.encoding = response.apparent_encoding
    csvdata = response.text.splitlines()
    return list(filter(any, list(csv.reader(csvdata))[1:]))


def google_sheet_to_csv_url(share_url):
    """
    Input: The share URL of a Google Sheets sheet.

    Output: A URL that downloads (exports) the sheet in CSV format.
    """
    match = re.search("/edit.*$", share_url)
    if not match:
        raise ValueError(f"{share_url} doesn't seem like a Google Sheets URL.")

    url = share_url[: match.start()]
    return url + "/export?format=csv"


def get_TAs_from_csv(csv_row):
    """
    Input: takes a CSV data row as from a csv.reader.

    Output: returns the list of signed TAs. Ensures casefold for TA IDs.
    """
    return list(
        map(
            lambda x: x.casefold(),
            filter(lambda x: x.strip(), csv_row[len(SIGNUP_SHEET_HEADER) :]),
        )
    )


def get_booked_TAs_from_csv(csv_row):
    """
    Input: takes a CSV data row as from a csv.reader.

    Output: returns the list of signed TAs, the first N, where N is the number of
    needed TAs specified in the CSV data.
    """
    TAs = get_TAs_from_csv(csv_row)
    num_needed_TAs = int(csv_row[SIGNUP_SHEET_HEADER.index("#Needed TAs")])

    return TAs[:num_needed_TAs], TAs[num_needed_TAs:]


def filter_events_by_TA(ta_id, csv_rows):
    """
    Input: ta_id is the string to (exactly) match the TAs' identifiers against;
    csv_rows is a list of CSV rows, as from csv.reader.

    Output: a list of CSV rows containing only the rows containing ta_id.
    """
    return list(filter(lambda x: ta_id.casefold() in get_TAs_from_csv(x), csv_rows))


def filter_events_by_title(event_title, csv_rows):
    """
    Input: event_title is the substring to match the event title against;
    csv_rows is a list of CSV rows, as from csv.reader.

    Output: a list of CSV rows containing only the rows with an event title
    having event_title as substring.
    """
    return list(
        filter(lambda x: event_title in x[SIGNUP_SHEET_HEADER.index("Event")], csv_rows)
    )



import arrow.arrow as arrow
import calendar
import dateutil
import ics.event
import ics.icalendar
import requests

TIMEEDIT_EVENTS = [
    "Datorlaboration",
    "Computer lab",
    "Examination",
    "Exkursion",
    "Excursion",
    "Fältövning",
    "Field exercise",
    "Frågestund",
    "Question time",
    "Question Time",
    "Föreläsning",
    "Lecture",
    "FX-komplettering",
    "FX-completion",
    "Handledning",
    "Tutorial",
    "Information",
    "Kontrollskrivning",
    "Partial Exam/Quiz",
    "Laboration",
    "Lab",
    "Lektion",
    "Lesson",
    "Omtenta",
    "Re-exam",
    "Re-Exam",
    "Presentation",
    "Projektarbete",
    "Project work",
    "Project Work",
    "Redovisning",
    "Reporting",
    "Räknestuga",
    "Math work session",
    "Seminarium",
    "Seminar",
    "Samverkansinlärning",
    "Supplemental instruction",
    "Supplemental Instruction",
    "Studiebesök",
    "Study visit",
    "Tentamen",
    "Exam",
    "Upprop",
    "Roll-call",
    "Webinarium",
    "Webinar",
    "Verksamhetsförlagd utbildning",
    "Pre-service Placement",
    "Workshop",
    "Övning",
    "Exercise",
]


def format_header(event):
    """
    Formats the event header, a one-line representation (week, date, time,
    event title)
    """
    header = (
        f"Week {event.begin.isocalendar()[1]} "
        + calendar.day_name[event.begin.weekday()]
        + " "
        + event.begin.to(dateutil.tz.tzlocal()).format("DD/MM HH:mm")
    )

    header += f" {event.name}"

    return header


def format_event(event):
    """
    Takes event (ics.event.Event) object,
    returns long string representation (header, location, description over
    several lines)
    """
    header = format_header(event)
    location = event.location
    description = "\n".join(
        filter(lambda x: "http" not in x, event.description.splitlines())
    )

    return f"{header}\n{location}\n{description}".strip()


def filter_event_description(
    event_desc, ignore=["http", "grupp", "group", "ID ", "Daniel Bosk"], separator="; "
):
    """
    Takes event description event_desc as string,
    returns filtered string with newlines replaced by semicolons.
    Rows of description containing any term in list ignore, will not appear in
    the returned string.
    The filtered rows of the description are joined again by the string in
    separator.
    """
    desc_parts = event_desc.splitlines()
    desc_parts_keep = []

    for part in desc_parts:
        found = False
        for term in ignore:
            if term in part:
                found = True
                break

        if not found:
            desc_parts_keep.append(part)

    return separator.join(desc_parts_keep)


def format_event_short(event):
    """Takes event (ics.event.Event) object,
    returns a short string representation (one line)"""
    header = format_header(event)
    description = filter_event_description(event.description)

    return f"{header} {description}".strip()


def read_calendar(url):
    """
    Input: url is a string containing the URL to the ICS-formatted calendar.
    Output: an [[ics.icalendar.Calendar]] object.
    """
    response = requests.get(url)
    if response.status_code == requests.codes.ok:
        return ics.icalendar.Calendar(imports=response.text)
    raise Exception(response.text)


def event_filter(events, whitelisted=TIMEEDIT_EVENTS):
    """
    Takes a list of events (ics.event.Event), returns a filtered list of events
    (generator). The events to include are the teaching events.

    It covers the following events:

      "Datorlaboration", "Computer lab",
      "Examination",
      "Exkursion", "Excursion",
      "Fältövning", "Field exercise",
      "Frågestund", "Question time", "Question Time",
      "Föreläsning", "Lecture",
      "FX-komplettering", "FX-completion",
      "Handledning", "Tutorial",
      "Information",
      "Kontrollskrivning", "Partial Exam/Quiz",
      "Laboration", "Lab",
      "Lektion", "Lesson",
      "Omtenta", "Re-exam", "Re-Exam",
      "Presentation",
      "Projektarbete", "Project work", "Project Work",
      "Redovisning", "Reporting",
      "Räknestuga", "Math work session",
      "Seminarium", "Seminar",
      "Samverkansinlärning", "Supplemental instruction",
      "Supplemental Instruction",
      "Studiebesök", "Study visit",
      "Tentamen", "Exam",
      "Upprop", "Roll-call",
      "Webinarium", "Webinar",
      "Verksamhetsförlagd utbildning", "Pre-service Placement",
      "Workshop",
      "Övning", "Exercise"
    """
    for event in events:
        for event_type in whitelisted:
            if event_type in event.name:
                yield event
                break



from config import COURSES, SIGNUP
from nytid.signup import sheets
from nytid.signup import hr

booked = []

for course, url in SIGNUP.items():
    booked += sheets.read_signup_sheet_from_url(url)

for user in hr.hours_per_TA(booked):
    print(user)




import datetime
from enum import Enum
import ics.icalendar
import logging
import typer
from typing_extensions import Annotated

from nytid.cli import courses as coursescli
from nytid import courses as courseutils
from nytid import schedules as schedutils

cli = typer.Typer(name="schedule", help="Working with course schedules")


def update_end_date(start, end):
    """
    Returns a correct end date.
    """
    if end < start:
        return start + datetime.timedelta(weeks=1)
    return end


start_date_opt = typer.Option(help="The start date", formats=["%Y-%m-%d"])
end_date_opt = typer.Option(help="The end date", formats=["%Y-%m-%d"])


class GroupByDayOrWeek(str, Enum):
    week = "week"
    day = "day"


group_by_day_or_week = typer.Option(
    help="Choose whether to group events " "by day or week", case_sensitive=False
)


@cli.command()
def show(
    course: Annotated[str, coursescli.course_arg_regex] = ".*",
    register: Annotated[str, coursescli.register_opt_regex] = coursescli.MINE,
    start: Annotated[datetime.datetime, start_date_opt] = str(datetime.date.today()),
    end: Annotated[datetime.datetime, end_date_opt] = str(
        datetime.date.today() + datetime.timedelta(weeks=1)
    ),
    group_by: Annotated[GroupByDayOrWeek, group_by_day_or_week] = "week",
):
    """
    Shows schedule for courses in human readable format.
    """
    end = update_end_date(start, end)
    registers = coursescli.registers_regex(register)
    the_courses = list(coursescli.courses_regex(course, registers))

    schedule = ics.icalendar.Calendar()

    for a_course, a_register in the_courses:
        try:
            course_conf = courseutils.get_course_config(a_course, a_register)
        except KeyError as err:
            logging.error(f"Can't find {a_course} in {a_register}: {err}")
            continue

        try:
            course_ics_url = course_conf.get("ics")
        except KeyError as err:
            logging.error(f"Can't find schedule URL for {a_course}: {err}")
            continue

        course_schedule = schedutils.read_calendar(course_ics_url)

        schedule.events.update(course_schedule.events)
    first = True
    if group_by == GroupByDayOrWeek.week:
        group_by_idx = 1
    elif group_by == GroupByDayOrWeek.day:
        group_by_idx = 2
    for event in schedule.timeline:
        if event.end.date() < start.date():
            continue
        elif event.begin.date() > end.date():
            continue
        if first:
            first = False
            current_epoc = event.begin.isocalendar()[group_by_idx]
        elif event.begin.isocalendar()[group_by_idx] != current_epoc:
            print("\n")
            current_epoc = event.begin.isocalendar()[group_by_idx]

        print(schedutils.format_event_short(event))



"""The CLI of nytid"""

import typer
from nytid.cli import courses
import typerconf as config
from nytid.cli import schedule
from nytid.cli import signupsheets

import logging
import sys

logging.basicConfig(format=f"nytid: %(levelname)s: %(message)s")

cli = typer.Typer()

cli.add_typer(courses.cli, name="courses")
config.add_config_cmd(cli)
cli.add_typer(schedule.cli, name="schedule")
cli.add_typer(signupsheets.cli, name="signupsheets")

if __name__ == "__main__":
    cli()
